import logging
from io import IOBase
from pathlib import PurePath
from typing import Optional, Union

from pdfminer.high_level import extract_pages

from hotpdf.memory_map import MemoryMap

logging.getLogger("pdfminer").setLevel(logging.ERROR)


def __process(
    source: Union[PurePath, str, IOBase],
    password: str = "",
    page_numbers: Optional[list[int]] = None,
) -> list[MemoryMap]:
    pages: list[MemoryMap] = []
    page_numbers = sorted(page_numbers) if page_numbers else []
    hl_page_layouts = extract_pages(source, password=password, page_numbers=page_numbers, caching=False)
    for page_layout in hl_page_layouts:
        parsed_page: MemoryMap = MemoryMap()
        parsed_page.build_memory_map()
        parsed_page.load_memory_map(page=page_layout)
        pages.append(parsed_page)
    return pages


def process(
    source: Union[PurePath, str, IOBase],
    password: str = "",
    page_numbers: Optional[list[int]] = None,
) -> list[MemoryMap]:
    return __process(
        source=source,
        password=password,
        page_numbers=page_numbers,
    )



import shutil

import pytest

from hotpdf import HotPdf
from hotpdf.data.classes import ElementDimension as El
from hotpdf.data.classes import HotCharacter
from hotpdf.sparse_matrix import SparseMatrix
from hotpdf.utils import filter_adjacent_coords, intersect, to_text


def xml_copy_file_name(xml_file_name: str):
    shutil.copy(xml_file_name, f"{xml_file_name}_copy.xml")
    return f"{xml_file_name}_copy.xml"


def test_load_file(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)


def test_element_dimensions_empty():
    assert filter_adjacent_coords(["d", "u", "m", "m", "y"], []) == []


def test_sparse_matrix_insert_and_get():
    matrix = SparseMatrix(3, 3)
    matrix.insert("A", 0, 0)
    matrix.insert("B", 1, 1)
    matrix.insert("C", 2, 2)
    assert matrix.get(0, 0) == "A"
    assert matrix.get(1, 1) == "B"
    assert matrix.get(2, 2) == "C"

    # Test getting non-inserted values (should return empty string)
    assert matrix.get(0, 1) == ""
    assert matrix.get(1, 0) == ""
    assert matrix.get(2, 1) == ""


def test_sparse_matrix_getitem_and_setitem():
    matrix = SparseMatrix(3, 3)
    matrix[0, 0] = "A"
    matrix[1, 1] = "B"
    matrix[2, 2] = "C"
    assert matrix[0, 0] == "A"
    assert matrix[1, 1] == "B"
    assert matrix[2, 2] == "C"

    matrix[1, 1] = "D"
    assert matrix[1, 1] == "D"
    with pytest.raises(IndexError, match="Specified index is out of range"):
        matrix[0, -1]
    with pytest.warns(UserWarning, match="Index Error. Skipping insertion into SparseMatrix"):
        matrix[-1, 0] = "Y"

    with pytest.raises(IndexError, match="Specified index is out of range"):
        matrix.get(-1, 0)
    with pytest.warns(UserWarning, match="Index Error. Skipping insertion into SparseMatrix"):
        matrix.insert("Y", -1, 0)


def test_sparse_matrix_iterator():
    matrix = SparseMatrix(3, 3)
    matrix[0, 0] = "A"
    matrix[1, 1] = "B"
    matrix[2, 2] = "C"
    non_empty_values = list(matrix)
    expected_result = [((0, 0), "A"), ((1, 1), "B"), ((2, 2), "C")]
    assert non_empty_values == expected_result


@pytest.mark.parametrize(
    "bbox1, bbox2, expected",
    [
        (El(0, 0, 2, 2), El(3, 3, 5, 5), False),
        (El(0, 0, 2, 2), El(3, 3, 4, 4), False),
        (El(0, 0, 2, 2), El(3, 3, 5, 5), False),
        (El(1, 1, 4, 4), El(3, 2, 6, 5), True),
        (El(1, 1, 6, 6), El(2, 2, 5, 5), True),
        (El(0, 0, 4, 4), El(4, 0, 8, 4), True),
        (El(0, 0, 3, 3), El(3, 3, 6, 6), True),
        (El(2, 1, 5, 4), El(4, 0, 7, 3), True),
        (El(1, 2, 4, 5), El(0, 4, 3, 7), True),
        (El(0, 0, 3, 3), El(0, 0, 3, 3), True),
        (El(2, 0, 6, 4), El(4, 2, 8, 6), True),
    ],
)
def test_intersect(bbox1, bbox2, expected):
    assert intersect(bbox1, bbox2) == expected


@pytest.mark.parametrize(
    "hot_characters, expected",
    [
        ([HotCharacter(value="H", x=0, y=0, x_end=10, span_id="x")], "H"),
        (
            [
                HotCharacter(value="H", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="e", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="l", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="l", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="o", x=0, y=0, x_end=10, span_id="x"),
            ],
            "Hello",
        ),
        (
            [
                HotCharacter(value="H", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="e", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="l", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="l", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="o", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value=" ", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="W", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="o", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="r", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="l", x=0, y=0, x_end=10, span_id="x"),
                HotCharacter(value="d", x=0, y=0, x_end=10, span_id="x"),
            ],
            "Hello World",
        ),
    ],
)
def test_to_text(hot_characters, expected):
    assert to_text(hot_characters) == expected


def test_invalid_page_number(valid_file_name):
    hotpdf_object = HotPdf()
    hotpdf_object.load(valid_file_name)
    with pytest.raises(ValueError, match="Invalid page number"):
        _ = hotpdf_object.extract_page_text(page=2)


def test_len_memory_map_magic_function(valid_file_name):
    hotpdf_object = HotPdf()
    hotpdf_object.load(valid_file_name)
    assert len(hotpdf_object.pages[0].span_map) > 10


def test_span_map_get_span_none(valid_file_name):
    hotpdf_object = HotPdf(valid_file_name)
    assert hotpdf_object.pages[0].span_map.get_span("xyz") is None
    assert hotpdf_object.pages[0].span_map["xyz"] is None


@pytest.mark.skip(reason="Not implemented")
def test_span_map_set_none_error(valid_file_name):
    # Setting non span object in spanmap should throw error
    pass



from collections import defaultdict

from .data.classes import HotCharacter, PageResult


class TrieNode:
    def __init__(self) -> None:
        """Initialize a TrieNode.

        Args:
            children (dict): Mapping of characters to TrieNode objects.
            is_end_of_word (bool): Flag indicating the end of a word.
            hot_characters (list[HotCharacter]): List of HotCharacter instances associated with the word.
        """
        self.children: defaultdict[str, TrieNode] = defaultdict(TrieNode)
        self.is_end_of_word: bool = False
        self.hot_characters: list[HotCharacter] = []


class Trie:
    def __init__(self) -> None:
        """Initialize a Trie."""
        self.root = TrieNode()

    def insert(self, word: str, hot_character: HotCharacter) -> None:
        """Insert a word and a HotCharacter into the Trie.

        Args:
            word (str): The word to insert.
            hot_character (HotCharacter): The HotCharacter to insert.
        """
        node: TrieNode = self.root
        for char in word:
            node = node.children[char]
        node.is_end_of_word = True
        node.hot_characters = node.hot_characters + [hot_character]

    def search_all(self, text: str) -> tuple[list[str], PageResult]:
        """Search for words in the Trie that match a given text.

        Args:
            text (str): The text to search for.

        Returns:
            tuple: A tuple containing a list of the target characters found and corresponding PageResult.
        """
        node: TrieNode = self.root
        found: list[str] = []
        hot_characters: PageResult = []
        for char in text:
            if char in node.children and node.children[char].is_end_of_word:
                found.append(char)
                if node.children[char].hot_characters:
                    hot_characters.append(node.children[char].hot_characters)
            node = self.root
        return found, hot_characters



from collections import defaultdict
from collections.abc import Iterator
from warnings import warn


class SparseMatrix:
    """2D representation of a PDF in plain text format.

    Sparse matrix removes the need to keep in memory the blank spaces
    that are present in a PDF. Thus reducing memory usage drastically.
    """

    def __init__(self, rows: int = 0, columns: int = 0):
        self.values: defaultdict[tuple[int, int], str] = defaultdict(str)
        self.rows = rows
        self.columns = columns

    def __getitem__(self, key: tuple[int, int]) -> str:
        row_idx, column_idx = key
        self.__check_indices(row_idx, column_idx)
        return self.values[(row_idx, column_idx)]

    def __update_indices(self, row_idx: int, column_idx: int) -> None:
        if row_idx > self.rows:
            self.rows = row_idx + 1
        if column_idx > self.columns:
            self.columns = column_idx + 1

    def __check_indices(self, row_idx: int, column_idx: int) -> None:
        if row_idx < 0 or column_idx < 0:
            raise IndexError("Specified index is out of range")

    def __setitem__(self, key: tuple[int, int], value: str) -> None:
        row_idx, column_idx = key
        try:
            self.__check_indices(row_idx, column_idx)
        except IndexError as ie:
            warn("Index Error. Skipping insertion into SparseMatrix", stacklevel=1)
            warn(str(ie), stacklevel=1)
            return
        self.__update_indices(row_idx, column_idx)
        if value:
            self.values[(row_idx, column_idx)] = value

    def __iter__(self) -> Iterator[tuple[tuple[int, int], str]]:
        yield from self.values.items()

    def insert(self, value: str, row_idx: int, column_idx: int) -> None:
        self.__update_indices(row_idx, column_idx)
        try:
            self.__check_indices(row_idx, column_idx)
        except IndexError:
            warn("Index Error. Skipping insertion into SparseMatrix", stacklevel=1)
            return
        if value:
            self.values[(row_idx, column_idx)] = value

    def get(self, row_idx: int, column_idx: int) -> str:
        self.__check_indices(row_idx, column_idx)
        return self.values[(row_idx, column_idx)]



from typing import Union

from .data.classes import ElementDimension, HotCharacter, PageResult


def find_neighbour_coord(
    reference_character: HotCharacter,
    hot_characters: list[HotCharacter],
    max_distance: int = 5,
    span_tolerance: int = 5,
) -> Union[HotCharacter, None]:
    """Find a neighbouring coordinate within a specified maximum distance.

    Args:
        reference_character (HotCharacter): Previous char instance.
        hot_characters (list[HotCharacter]): List of character instances to search through.
        max_distance (int): Maximum distance between reference coordinate and target coordinate.
        span_tolerance (int): Additional distance to consider if text lies in the same span.

    Returns:
        HotCharacter: The neighbouring HotCharacter if found, else None.
    """
    for hot_character in hot_characters:
        if hot_character == reference_character:
            continue
        if (
            (
                0
                <= (hot_character.x - reference_character.x_end)
                <= (
                    (max_distance + span_tolerance)
                    if hot_character.span_id == reference_character.span_id
                    and (hot_character.span_id and reference_character.span_id)
                    else max_distance
                )
            )
            or (reference_character.span_id == hot_character.span_id and reference_character.x < hot_character.x)
            and (hot_character.x <= reference_character.x_end)
        ) and reference_character.y == hot_character.y:
            return hot_character
    return None


def filter_adjacent_coords(text: list[str], page_hot_character_occurences: PageResult) -> PageResult:
    """Filter adjacent coordinates based on the given text.

    Args:
        text (str): The text to filter by.
        page_hot_character_occurences (list): List of coordinates to filter by page

    Returns:
        PageResult: List of adjacent groups of HotCharacters on a page.
    """
    if not page_hot_character_occurences:
        return []

    max_len = len(text)
    adjacent_groups = []

    anchor_hot_character_instances = page_hot_character_occurences[0]

    for anchor_hot_character in anchor_hot_character_instances:
        neighbours = [anchor_hot_character]
        reference_hot_character = anchor_hot_character
        for coords_j in page_hot_character_occurences[1:]:
            neighbour_hot_character = find_neighbour_coord(
                reference_character=reference_hot_character,
                hot_characters=coords_j,
            )
            if neighbour_hot_character:
                neighbours.append(neighbour_hot_character)
                reference_hot_character = neighbour_hot_character
        if len(neighbours) == max_len:
            adjacent_groups.append(neighbours[:])
            neighbours.clear()
        neighbours = []
    return adjacent_groups


def get_element_dimension(elem: list[HotCharacter]) -> ElementDimension:
    """Get the dimensions of an element based on its coordinates.

    Args:
        elem (list): List of coordinates representing an element.

    Returns:
        ElementDimension: ElementDimension object containing the dimensions (x0, x1, y0, y1, span_id).
    """
    x0 = min(elem, key=lambda item: item.x).x
    x1 = max(elem, key=lambda item: item.x_end).x_end
    y0 = min(elem, key=lambda item: item.y).y
    y1 = max(elem, key=lambda item: item.y).y
    span = elem[0].span_id
    return ElementDimension(x0=x0, x1=x1, y0=y0, y1=y1, span_id=span)


def intersect(bbox1: ElementDimension, bbox2: ElementDimension) -> bool:
    """Check if two bounding boxes intersect.

    Args:
        bbox1 (ElementDimension): Bounding box 1. (x0, y0, x1, y1)
        bbox2 (ElementDimension): Bounding box 2. (x0, y0, x1, y1)

    Returns:
        bool: True if the bounding boxes intersect, else False.
    """
    return not (bbox2.x0 > bbox1.x1 or bbox2.x1 < bbox1.x0 or bbox2.y0 > bbox1.y1 or bbox2.y1 < bbox1.y0)


def to_text(el: list[HotCharacter]) -> str:
    """Convert a list of HotCharacter instances to text.

    Args:
        el (list): List of HotCharacter instances.

    Returns:
        str: The text.
    """
    return "".join(char.value for char in el)



import math
from collections.abc import Generator

from pdfminer.layout import LTAnno, LTChar, LTComponent, LTPage, LTText, LTTextContainer, LTTextLine

from .data.classes import HotCharacter, PageResult
from .helpers.nanoid import generate_nano_id
from .span_map import SpanMap
from .sparse_matrix import SparseMatrix
from .trie import Trie


class MemoryMap:
    def __init__(self) -> None:
        """Initialize the MemoryMap. 2D Matrix representation of a PDF Page.

        Args:
            width (int): The width (max columns) of a page.
            height (int) The height (max rows) of a page.
        """
        self.text_trie = Trie()
        self.span_map = SpanMap()
        self.width: int = 0
        self.height: int = 0

    def build_memory_map(self) -> None:
        """Build the memory map based on width and height.

        The memory map is a SparseMatrix representation of the PDF.
        """
        self.memory_map = SparseMatrix()

    def __reverse_page_objs(self, page_objs: list[LTComponent]) -> Generator[LTComponent, None, None]:
        yield from reversed(page_objs)

    def __get_page_spans(self, page: LTPage) -> Generator[LTTextLine, None, None]:
        element_stack = self.__reverse_page_objs(page._objs)
        for obj in element_stack:
            if isinstance(obj, LTTextLine):
                yield obj
            elif isinstance(obj, LTTextContainer):
                element_stack = self.__reverse_page_objs(list(obj))
                yield from (em for em in element_stack if isinstance(em, LTTextLine))

    def load_memory_map(self, page: LTPage) -> None:
        """Load memory map data from an XML page.

        Args:
            page (str): LTPage Element returned by pdfminer

        Returns:
            None
        """
        char_hot_characters: list[tuple[str, HotCharacter]] = []
        spans: Generator[LTTextLine, None, None] = self.__get_page_spans(page)
        for span in spans:
            span_id = generate_nano_id(size=10)
            prev_char_inserted = False
            for character in span:
                if isinstance(character, (LTChar, LTText)) and (
                    hasattr(character, "x0")
                    and hasattr(character, "x1")
                    and hasattr(character, "y0")
                    and hasattr(character, "y1")
                ):
                    char_c = character.get_text()
                    x0 = round(character.x0)
                    x1 = round(character.x1)
                    y0 = round(page.height - character.y0)
                    hot_character: HotCharacter = self.__insert_element_memory_map(
                        value=char_c, x=x0, y=y0, x_end=x1, span_id=span_id
                    )
                    char_hot_characters.append((
                        char_c,
                        hot_character,
                    ))
                    prev_char_inserted = char_c != " "
                elif isinstance(character, LTAnno) and (character._text == " ") and prev_char_inserted:
                    _elem_width: int = 1
                    space_char: HotCharacter = self.__insert_element_memory_map(
                        value=" ", x=x0, y=y0, x_end=x0 + _elem_width, span_id=span_id
                    )
                    char_hot_characters.append((
                        " ",
                        space_char,
                    ))
                    prev_char_inserted = False

        # Insert into Trie and Span Maps
        _hot_character: HotCharacter
        for char_c, _hot_character in char_hot_characters:
            self.text_trie.insert(char_c, _hot_character)
            if _hot_character.span_id:
                self.span_map[_hot_character.span_id] = _hot_character
        self.width = math.ceil(page.width)
        self.height = math.ceil(page.height)

    def __insert_element_memory_map(self, value: str, x: int, y: int, x_end: int, span_id: str) -> HotCharacter:
        """Insert element into memory map.

        Args:
            value (str): Value of the object to be inserted.
            x (int): column index (x0-coordinate) of the element.
            y (int): row index (y0-coordinate) of the element.
            x_end (int): end column index (x1-coordinate) of element. x_end - x = width of element.
            span_id (str): id of parent span.

        Returns:
            HotCharacter: HotCharacter object of the whitespace.
        """
        self.memory_map.insert(value=value, row_idx=y, column_idx=x)
        return HotCharacter(
            value=value,
            x=x,
            y=y,
            x_end=x_end,
            span_id=span_id,
        )

    def extract_text_from_bbox(self, x0: int, x1: int, y0: int, y1: int) -> str:
        """Extract text within a specified bounding box.

        Args:
            x0 (int): Left x-coordinate of the bounding box.
            x1 (int): Right x-coordinate of the bounding box.
            y0 (int): Bottom y-coordinate of the bounding box.
            y1 (int): Top y-coordinate of the bounding box.

        Returns:
            str: Extracted text within the bounding box.
        """
        extracted_text: str = ""
        for row in range(max(y0, 0), min(y1, self.memory_map.rows - 1) + 1):
            row_text: str = ""
            row_text = "".join(
                self.memory_map.get(row_idx=row, column_idx=col)
                for col in range(max(x0, 0), min(x1, self.memory_map.columns - 1) + 1)
            )
            if row_text:
                extracted_text += row_text + "\n"

        return extracted_text

    def find_text(self, query: str) -> tuple[list[str], PageResult]:
        """Find text within the memory map.

        Args:
            query (str): The text to search for.

        Returns:
            list: List of found text coordinates.
        """
        found_text = self.text_trie.search_all(query)
        return found_text



from collections import defaultdict
from dataclasses import dataclass
from typing import Optional


@dataclass
class HotCharacter:
    """A hot character is a character on a page with certain attributes.

    Attributes:
        value (str): value value of the character.
        x (int): x position of the character - column number.
        y (int): y position of the character - row number.
        x_end (int): end x position of the character. x_end - x = width of character.
        span_id (str, Optional): hash of parent span the character lies in.
    """

    value: str
    x: int
    y: int
    x_end: int
    span_id: str


@dataclass
class ElementDimension:
    """ElementDimension is the dimension of an element in hotpdf.

    Attributes:
        x0 (int): starting x position of the element (column).
        y0 (int): starting y position of the element (row).
        x1 (int): end x position of the element (column). x1 - x0 = width.
        y1 (int): end y position of the element (row) y1 - y0 = height.
        span_id (str, Optional): hash of parent span the element lies in.
    """

    x0: int
    y0: int
    x1: int
    y1: int
    span_id: Optional[str] = None


@dataclass(init=True)
class Span:
    """A span is a group of characters that are close to each other.

    Attributes:
        characters (list[HotCharacter]): list of characters in the span.
        x0 (int): starting x position of the span (column).
        y0 (int): starting y position of the span (row).
        x_end (int): end x position of the span (column). x_end - x0 = width.
        span_id (str, Optional): hash of the span.
    """

    characters: list[HotCharacter]
    span_id: str

    def to_text(self) -> str:
        """Convert the span to text.

        Returns:
            str: text representation of the span.
        """
        return "".join(char.value for char in self.characters)

    def get_element_dimension(self) -> ElementDimension:
        """Get the element dimension of the span.

        Raises:
            ValueError: if the span has no characters.

        Returns:
            ElementDimension: _description_
        """
        if not self.characters:
            raise ValueError("Span has no characters")
        x0 = self.characters[0].x
        y0 = self.characters[0].y
        x1 = self.characters[-1].x_end
        y1 = self.characters[-1].y
        return ElementDimension(x0, y0, x1, y1, self.span_id)


# All occurences of HotCharacters in a page
# A list[HotCharacter] is the representation of a word split into "HotCharacters"
# A list[list[HotCharacter]] is a list of multiple list[HotCharacter] found on a page
PageResult = list[list[HotCharacter]]

# Complete PageResult with Page Number as the index
SearchResult = defaultdict[int, PageResult]
